/*
 * Copyright (c) 2021 NIBIO <http://www.nibio.no/>. 
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 */

package no.nibio.model.maizepheno;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.TimeZone;
import java.util.stream.Collectors;
import no.nibio.vips.entity.ModelConfiguration;
import no.nibio.vips.entity.Result;
import no.nibio.vips.entity.ResultImpl;
import no.nibio.vips.entity.WeatherObservation;
import no.nibio.vips.i18n.I18nImpl;
import no.nibio.vips.model.ConfigValidationException;
import no.nibio.vips.model.Model;
import no.nibio.vips.model.ModelExcecutionException;
import no.nibio.vips.model.ModelId;
import no.nibio.vips.util.CommonNamespaces;
import no.nibio.vips.util.InvalidAggregationTypeException;
import no.nibio.vips.util.ModelUtil;
import no.nibio.vips.util.WeatherElements;
import no.nibio.vips.util.WeatherObservationListException;
import no.nibio.vips.util.WeatherUtil;


/**
 * @copyright 2022 <a href="http://www.nibio.no/">NIBIO</a>
 * @author Tor-Einar Skog <tor-einar.skog@nibio.no>
 */
public class MaizePhenologyModel extends I18nImpl implements Model{

    public final static ModelId MODEL_ID = new ModelId("MAIZEPHENO");
    
    private Date sowingDate;
    private TimeZone timeZone;
    
    private Double baseTemp = 10.0; // TODO: Make configurable
    
    private PhaseInfo phaseInfo;
    
    // Helper class
    private final ModelUtil modelUtil;
    private final WeatherUtil weatherUtil;
    private DataMatrix dataMatrix;
    
    public MaizePhenologyModel()
    {
        super("no.nibio.vips.model.maizepheno.texts");
        this.modelUtil = new ModelUtil();
        this.weatherUtil = new WeatherUtil();
    }
    
    @Override
    public List<Result> getResult() throws ModelExcecutionException {
        // Start at sowing date
        
        Date currentDate = this.sowingDate;
        Calendar cal = Calendar.getInstance(this.timeZone);
        Date lastDate = this.dataMatrix.getLastDateWithParameterValue(DataMatrix.TND);
        Double heatSum = 0.0;
        List<Result> retVal = new ArrayList<>();

        DecimalFormatSymbols symbols = new DecimalFormatSymbols(Locale.getDefault());
        symbols.setDecimalSeparator('.');
        DecimalFormat dFormat = new DecimalFormat("###.##", symbols);
        while(!currentDate.after(lastDate))
        {
            Result result = new ResultImpl();
            result.setValidTimeStart(currentDate);
            Double TND = this.dataMatrix.getParamDoubleValueForDate(currentDate, DataMatrix.TND);
            Double TXD = this.dataMatrix.getParamDoubleValueForDate(currentDate, DataMatrix.TXD);
            Double dailyContribution = (TND != null && TXD != null) ? this.getDailyContribution(TND,TXD) // Default: All is well
                    : retVal.size() > 0 ? Double.valueOf(retVal.get(retVal.size()-1).getValue(MaizePhenologyModel.MODEL_ID.toString(), DataMatrix.DAILY_HEAT_SUM_CONTRIBUTION)) // Simple copy from yesterday
                    : 0.0; // Panic
            heatSum += dailyContribution != null ? dailyContribution : 0.0;
            
            result.setValue(CommonNamespaces.NS_WEATHER, DataMatrix.TND, TND != null ? dFormat.format(TND) : "null");
            result.setValue(CommonNamespaces.NS_WEATHER, DataMatrix.TXD, TXD != null ? dFormat.format(TXD) : "null");
            result.setValue(MaizePhenologyModel.MODEL_ID.toString(), DataMatrix.DAILY_HEAT_SUM_CONTRIBUTION, dailyContribution != null ? dFormat.format(dailyContribution) : "null");
            result.setValue(MaizePhenologyModel.MODEL_ID.toString(), DataMatrix.HEAT_SUM, dFormat.format(heatSum));
            result.setValue(MaizePhenologyModel.MODEL_ID.toString(), DataMatrix.PHASE, this.phaseInfo.getPhase(heatSum).getPhaseName());
            
            retVal.add(result);
            
            // Moving on...
            cal.setTime(currentDate);
            cal.add(Calendar.DATE, 1);
            currentDate = cal.getTime();
        }
        return retVal;
    }
    
    private Double getDailyContribution(Double TND, Double TXD)
    {
        if(TND == null || TXD == null)
        {
            return null;
        }
        
        Double sortOfMean = (TND + TXD) / 2;
        return  sortOfMean < this.baseTemp ? 0.0 : sortOfMean - this.baseTemp;
        
    }

    @Override
    public ModelId getModelId() {
        return MaizePhenologyModel.MODEL_ID;
    }

    @Override
    public String getModelName() {
        return this.getModelName(Model.DEFAULT_LANGUAGE);
    }

    @Override
    public String getModelName(String language) {
        return this.getText("name", language);
    }

    @Override
    public String getLicense() {
        return 
                "  Copyright (c) 2021 NIBIO <http://www.nibio.no/>. \n" +
                " \n" +
                "  This program is free software: you can redistribute it and/or modify\n" +
                "  it under the terms of the GNU Affero General Public License as published by\n" +
                "  the Free Software Foundation, either version 3 of the License, or\n" +
                "  (at your option) any later version.\n" +
                " \n" +
                "  This program is distributed in the hope that it will be useful,\n" +
                "  but WITHOUT ANY WARRANTY; without even the implied warranty of\n" +
                "  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n" +
                "  GNU Affero General Public License for more details.\n" +
                " \n" +
                "  You should have received a copy of the GNU Affero General Public License\n" +
                "  along with this program.  If not, see <https://www.gnu.org/licenses/>.\n" +
                " \n" ;
    
    }

    @Override
    public String getCopyright() {
        return "(c) 2022 NIBIO (http://www.nibio.no/). Contact: post@nibio.no";
    }

    @Override
    public String getModelDescription() {
        return this.getModelDescription(Model.DEFAULT_LANGUAGE);
    }

    @Override
    public String getModelDescription(String language) {
        return this.getText("description", language);
    }

    @Override
    public String getWarningStatusInterpretation() {
        return this.getWarningStatusInterpretation(Model.DEFAULT_LANGUAGE);
    }

    @Override
    public String getWarningStatusInterpretation(String language) {
        return this.getText("statusInterpretation", language);
    }

    @Override
    public String getModelUsage() {
        return this.getModelUsage(Model.DEFAULT_LANGUAGE);
    }

    @Override
    public String getModelUsage(String language) {
        return this.getText("usage", language);
    }

    @Override
    public String getSampleConfig() {
        return "{\n" +
"	\"sowingDate\":\"2020-11-18\",\n" +
"	\"timeZone\":\"GMT+01:00\",\n" +
"	\"observations\":[\n" +
"		{\"elementMeasurementTypeId\":\"TM\",\"logIntervalId\":1,\"timeMeasured\":\"2020-11-16T07:00:00Z\",\"value\":28.799999999999997},\n" +
"		{\"elementMeasurementTypeId\":\"TM\",\"logIntervalId\":1,\"timeMeasured\":\"2020-11-16T08:00:00Z\",\"value\":30.191666666666663},\n" +
"		{\"elementMeasurementTypeId\":\"TM\",\"logIntervalId\":1,\"timeMeasured\":\"2020-11-16T09:00:00Z\",\"value\":32.18333333333333}\n" +
"		[...]\n" +
"	]\n" +
"}";
    }

    @Override
    public void setConfiguration(ModelConfiguration config) throws ConfigValidationException {
        if(config.getConfigParameter("timeZone") == null)
        {
            throw new ConfigValidationException("ERROR: Time zone not specified");
        }
        else
        {
            // Setting timezone
            this.timeZone = TimeZone.getTimeZone((String) config.getConfigParameter("timeZone"));
        }
        
        if(config.getConfigParameter("sowingDate") == null)
        {
            throw new ConfigValidationException("ERROR: Missing sowing date");
        }
        else
        {
            SimpleDateFormat format = new SimpleDateFormat("yyyy-MM-dd");
            format.setTimeZone(timeZone);
            try
            {
                this.sowingDate = format.parse((String) config.getConfigParameter("sowingDate"));
            }
            catch(ParseException ex)
            {
                throw new ConfigValidationException("ERROR with sowing date (" + (String) config.getConfigParameter("sowingDate") + ": Not a valid date on format yyyy-MM-dd");
            }
        }
        List<WeatherObservation> inputWeatherData = this.modelUtil.extractWeatherObservationList(config.getConfigParameter("observations"));
        
        
        
        // Do we have any data?
        if(inputWeatherData == null || inputWeatherData.isEmpty())
        {
            throw new ConfigValidationException("ERROR: No weather data provided");
        }
        
        // Hourly data? Aggregate
        if(inputWeatherData.get(0).getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1H))
        {
            Set<String> parametersFound = inputWeatherData.stream().map(o->o.getElementMeasurementTypeId()).collect(Collectors.toSet());
            try
            {
                List<WeatherObservation> onlyTemps = this.weatherUtil.filterWeatherObservationsByParameter(inputWeatherData, parametersFound.contains(WeatherElements.TEMPERATURE_MINIMUM) ? WeatherElements.TEMPERATURE_MINIMUM : WeatherElements.TEMPERATURE_MEAN);
                List<WeatherObservation> minTemps = this.weatherUtil.getAggregatedDailyValues(onlyTemps, timeZone, 15, WeatherUtil.AGGREGATION_TYPE_MINIMUM,1, true);
                // If we only have mean temps, rename them to min temps
                if(minTemps.get(0).getElementMeasurementTypeId().equals(WeatherElements.TEMPERATURE_MEAN))
                {
                    this.weatherUtil.renameParameter(minTemps, WeatherElements.TEMPERATURE_MEAN, WeatherElements.TEMPERATURE_MINIMUM);
                }
                onlyTemps = this.weatherUtil.filterWeatherObservationsByParameter(inputWeatherData, parametersFound.contains(WeatherElements.TEMPERATURE_MAXIMUM) ? WeatherElements.TEMPERATURE_MAXIMUM : WeatherElements.TEMPERATURE_MEAN);
                List<WeatherObservation> maxTemps = this.weatherUtil.getAggregatedDailyValues(onlyTemps, timeZone, 15, WeatherUtil.AGGREGATION_TYPE_MAXIMUM,1, true);
                // If we only have mean temps, rename them to max temps
                if(maxTemps.get(0).getElementMeasurementTypeId().equals(WeatherElements.TEMPERATURE_MEAN))
                {
                    this.weatherUtil.renameParameter(maxTemps, WeatherElements.TEMPERATURE_MEAN, WeatherElements.TEMPERATURE_MAXIMUM);
                }
                inputWeatherData = minTemps;
                inputWeatherData.addAll(maxTemps);
            }
            catch(WeatherObservationListException | InvalidAggregationTypeException ex)
            {
                throw new ConfigValidationException("ERROR with weather data aggregation: " + ex.getMessage());
            }
        }
        else
        {
            List<String> requiredParameters = List.of(WeatherElements.TEMPERATURE_MINIMUM, WeatherElements.TEMPERATURE_MAXIMUM);
            // Check that we have daily min and max temps
            Set<String> parametersFound = new HashSet<>();
            for(WeatherObservation o:inputWeatherData)
            {
                if(o.getLogIntervalId().equals(WeatherObservation.LOG_INTERVAL_ID_1D))
                {
                    parametersFound.add(o.getElementMeasurementTypeId());
                }
            }
            if(! parametersFound.containsAll(requiredParameters))
            {
                throw new ConfigValidationException("ERROR with weather data: Missing daily values for minimum and/or maximum temperatures");
            }
        }
        
        this.dataMatrix = new DataMatrix();
        Collections.sort(inputWeatherData);
        
        /*
        ObjectMapper om = new ObjectMapper();
        try {
            System.out.println(om.writeValueAsString(inputWeatherData));
        } catch (JsonProcessingException ex) {
            Logger.getLogger(MaizePhenologyModel.class.getName()).log(Level.SEVERE, null, ex);
        }*/
        
        
        for(WeatherObservation o: inputWeatherData)
        {
            if(o.getElementMeasurementTypeId().equals(WeatherElements.TEMPERATURE_MINIMUM))
            {
                //System.out.println(o);
                this.dataMatrix.setParamDoubleValueForDate(o.getTimeMeasured(), DataMatrix.TND, o.getValue());
            }
            if(o.getElementMeasurementTypeId().equals(WeatherElements.TEMPERATURE_MAXIMUM))
            {
                //System.out.println(o);
                this.dataMatrix.setParamDoubleValueForDate(o.getTimeMeasured(), DataMatrix.TXD, o.getValue());
            }
        }
        
        // Check that sowing date is not before first daily temp
        if(this.sowingDate.before(this.dataMatrix.getFirstDateWithParameterValue(DataMatrix.TND)))
        {
            throw new ConfigValidationException("ERROR: sowing date (" + this.sowingDate + ") is before weather data starts (" + this.dataMatrix.getFirstDateWithParameterValue(DataMatrix.TND) + ")");
        }
        
        if(config.getConfigParameter("phaseInfo") == null)
        {
            this.phaseInfo = this.getDefaultPhaseInfo();
        }
        else
        {
            throw new ConfigValidationException("ERROR: You have provided phase info, but the model hasn't yet implemented how to import it");
        }
    }

    /**
     * @return 
     */
    private PhaseInfo getDefaultPhaseInfo() {
        String phaseInfoJson = "[\n" +
            "    {\n" +
            "        \"phaseName\": \"VE\",\n" +
            "        \"heatReq\"  : \"66.67\"\n" +
            "    },\n" +
            "    {\n" +
            "        \"phaseName\": \"V2\",\n" +
            "         \"heatReq\"  : \"111.11\"\n" +
            "    },\n" +
            "    {\n" +
            "        \"phaseName\": \"V3\",\n" +
            "        \"heatReq\"  : \"194.44\"\n" +
            "    },\n" +
            "    {\n" +
            "        \"phaseName\": \"V4-V6\",\n" +
            "        \"heatReq\"  : \"263.89\"\n" +
            "    }, \n" +
            "    {\n" +
            "        \"phaseName\": \"V7-V9\",\n" +
            "        \"heatReq\"  : \"338.89\"\n" +
            "    }, \n" +
            "    {\n" +
            "        \"phaseName\": \"V10\",\n" +
            "        \"heatReq\"  : \"411.11\"\n" +
            "    }, \n" +
            "    {\n" +
            "        \"phaseName\": \"VT\",\n" +
            "        \"heatReq\"  : \"630.55\"\n" +
            "    }, \n" +
            "    {\n" +
            "        \"phaseName\": \"R2\",\n" +
            "        \"heatReq\"  : \"922.22\"\n" +
            "    }, \n" +
            "    {\n" +
            "        \"phaseName\": \"R5\",\n" +
            "        \"heatReq\"  : \"1361.11\"\n" +
            "    }, \n" +
            "    {\n" +
            "        \"phaseName\": \"R6\",\n" +
            "        \"heatReq\"  : \"1500\"\n" +
            "    }\n" +
            "\n" +
            "]";
        
        PhaseInfo retVal = new PhaseInfo(phaseInfoJson);
        return retVal;
    }

}
